﻿using System.Threading;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Core.Networks;
using HeuristicLab.Data;
using HeuristicLab.Encodings.RealVectorEncoding;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Problems.DataAnalysis;

namespace HeuristicLab.Networks.IntegratedOptimization.SurrogateModeling {
  [Item("Full Evaluator", "")]
  [StorableClass]
  public sealed class FullEvaluatorNode : OrchestratedAlgorithmNode {
    public new FullEvaluationAlgorithm Algorithm {
      get { return (FullEvaluationAlgorithm)base.Algorithm; }
      set { base.Algorithm = value; }
    }

    #region Port Names
    private const string ExploitationFullEvalRequestPortName = "ExploitationFullEvalRequestPort";
    private const string ExplorationFullEvalRequestPortName = "ExplorationFullEvalRequestPort";
    private const string UpdateRegressionSolutionPortName = "Update RegressionSolution";
    #endregion Port Names

    #region Ports
    public IMessagePort ExploitationFullEvalRequestPort { get { return (IMessagePort)Ports[ExploitationFullEvalRequestPortName]; } }
    public IMessagePort ExplorationFullEvalRequestPort { get { return (IMessagePort)Ports[ExplorationFullEvalRequestPortName]; } }
    public IConfigurationPort UpdateRegressionSolutionPort { get { return (IConfigurationPort)Ports[UpdateRegressionSolutionPortName]; } }
    #endregion Ports

    #region Constructors & Cloning
    [StorableConstructor]
    private FullEvaluatorNode(bool deserializing) : base(deserializing) { }
    private FullEvaluatorNode(FullEvaluatorNode original, Cloner cloner) : base(original, cloner) {
      RegisterPortEvents();
    }
    public FullEvaluatorNode() : this("Full Evaluator") { }
    public FullEvaluatorNode(string name) : base(name) {
      var exploitationPort = CreateEvaluationPort<RealVector>(ExploitationFullEvalRequestPortName, "RealVector", "Quality");
      Ports.Add(exploitationPort);

      var explorationPort = CreateEvaluationPort<RealVector>(ExplorationFullEvalRequestPortName, "RealVector", "Quality");
      Ports.Add(explorationPort);

      var updatePort = new ConfigurationPort(UpdateRegressionSolutionPortName) {
        Parameters = {
          new PortParameter<IRegressionSolution>("RegressionSolution") {
            Type = PortParameterType.Input
          }
        }
      };
      Ports.Add(updatePort);

      Algorithm = new FullEvaluationAlgorithm();

      RegisterPortEvents();
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new FullEvaluatorNode(this, cloner);
    }
    #endregion Constructors & Cloning

    [StorableHook(HookType.AfterDeserialization)]
    private void AfterDeserialization() {
      RegisterPortEvents();
    }

    #region Helpers
    private void RegisterPortEvents() {
      ExploitationFullEvalRequestPort.MessageReceived += ExploitationFullEvalRequestPort_MessageReceived;
      ExplorationFullEvalRequestPort.MessageReceived += ExplorationFullEvalRequestPort_MessageReceived;
    }

    private void ExploitationFullEvalRequestPort_MessageReceived(object sender, EventArgs<IMessage, CancellationToken> e) {
      var evalMsg = e.Value;
      var realVector = (RealVector)evalMsg["RealVector"];
      var quality = (DoubleValue)evalMsg["Quality"];
      Algorithm.EnqueueForExploitation(realVector, quality.Value);
    }

    private void ExplorationFullEvalRequestPort_MessageReceived(object sender, EventArgs<IMessage, CancellationToken> e) {
      var evalMsg = e.Value;
      var realVector = (RealVector)evalMsg["RealVector"];
      var quality = (DoubleValue)evalMsg["Quality"];
      Algorithm.EnqueueForExploration(realVector, quality.Value);
    }
    #endregion Helpers
  }
}
